/*
  urldecode.c
   1  2006-03-18  Matlib
   2  2009-02-23  Matlib  + link against libarguments
                          + error handling
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <arguments.h>

#define BUFFER_SIZE 128

#define CHECK_ERROR(__COND__,__FUNC__) \
  { \
    if (__COND__) \
    { \
      perror (__FUNC__); \
      exit (EXIT_FAILURE); \
    } \
  }


struct Argument arguments [] =
{
  { 'f', "force", argument_bool, ARGUMENT_F_ONCE, NULL },
  { }
};
#define ARG_NFC (arguments [0].value == NULL)




typedef int EndCheck (char , unsigned int , unsigned int );

int end_zero (char c, unsigned int p, unsigned int l)
{
  return c == 0;
}

int end_length (char c, unsigned int p, unsigned int l)
{
  return p == l;
}

int mode;
char hex, c;
unsigned int hexval, p, k, ptot;

char digits [] = { '0', '1', '2', '3', '4', '5', '6', '7',
                   '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

char *allowed = "!*$,/?%;:@&0123456789-_.ABCDEFGHIJKLMNOPQRSTUVWXYZ#[]()'=+abcdefghijklmnopqrstuvwxyz~";
    
#define CHAR_CHECK(__COND__,__FALLBACK__,...) \
  { \
    if (__COND__) \
    { \
      fprintf (stderr, ARG_NFC ? "Error: " : "Warning: "); \
      fprintf (stderr, __VA_ARGS__); \
      if (ARG_NFC) \
        exit (EXIT_FAILURE); \
      else \
      { \
        __FALLBACK__ \
      } \
    } \
  } 
    
void decode (char *string, int len, EndCheck ec)
{
  p = 0;
loop:
  if (ec (string [p], p, len))
  {
    ptot += p;
    return;
  }
  for (k = 0; allowed [k] != string [p]; k ++)
    CHAR_CHECK (allowed [k] == 0, break;,
       "invalid character 0x%02X at position %d\n", string [p], p + ptot);
  if (mode == 0)
    switch (string [p])
    {
      case '%':
        mode ++;
        hexval = 0;
        break;
      case '+':
        putchar (' ');
        break;
      default:
        putchar (string [p]);
    }
  else
  {
    if (string [p] < 'a')
      c = string [p];
    else
      c = string [p] - 32;
    for (k = 0; digits [k] != c; k ++)
      CHAR_CHECK (digits [k] == 0,
         putchar ('%'); if (mode == 2) putchar (hex); 
         mode = 0; goto loop;,
         "character 0x%02X is not a hexadecimal digit at position %d\n", 
           string [p], p + ptot);
    hex = string [p];
    hexval += k << (4 * (2 - mode));
    if (mode == 2)
    {
      putchar (hexval);
      mode = 0;
    }
    else
      mode ++;
  }
  p ++;
  goto loop;
}


int main (int argc, char **argv)
{
  int i;
  char buffer [128];
  
  i = arguments_parse ("urldecode", "2 (2009-02-23)", "[url] ...", 
     "Converts a URL encoded string to plain text.", argc, argv, arguments);
  
  mode = 0;
  if (i == argc)
  {
loop:
    i = read (STDIN_FILENO, buffer, sizeof (buffer));
    CHECK_ERROR (i == -1, "read");
    if (i == 0)
      return EXIT_SUCCESS;
    decode (buffer, i, end_length);
    goto loop;
  }
  
  
  decode (argv [i ++], 0, end_zero);
  if (i < argc)
  {  
    CHECK_ERROR (putchar ('\n') == EOF, "putchar");
    do
    {
      decode (argv [i], 0, end_zero);
      CHECK_ERROR (putchar ('\n') == EOF, "putchar");
      i ++;
    }
    while (i < argc);
  }
  return EXIT_SUCCESS;
}
